(define-module (fruitflats sdl-private events)
  :use-module (fruitflats sdl-private common)
  :use-module (system foreign)
  :export (poll-event
           event-type
           event-pointer->vector))

;; Constants
(define event-symbols (make-vector (1+ 32)))
(let ((first-symbols  #(no-event
                        active-event
                        key-down
                        key-up
                        mouse-motion
                        mouse-button-down
                        mouse-button-up
                        joy-axis-motion
                        joy-ball-motion
                        joy-hat-motion
                        joy-button-down
                        joy-button-up
                        quit
                        sys-wm-event
                        event-reserved-a
                        event-reserved-b
                        video-resize
                        video-expose
                        event-reserved-2
                        event-reserved-3
                        event-reserved-4
                        event-reserved-5
                        event-reserved-6
                        event-reserved-7)))
  (vector-move-left! first-symbols
                     0 (vector-length first-symbols)
                     event-symbols 0)
  (vector-set! event-symbols 24 'user-event)
  (vector-set! event-symbols 32 'num-events))

(define event-field-types-alist `((active-event . (,uint8 ,uint8 ,uint8))
                                  (key-down . (,uint8 ,uint8 ,uint8))
                                  (key-up . (,uint8 ,uint8 ,uint8))
                                  (mouse-motion . (,uint8 ,uint8 ,uint8 ,uint16 ,int16 ,int16))
                                  (mouse-button-down . (,uint8 ,uint8 ,uint8 ,uint8 ,uint16))
                                  (mouse-button-up . (,uint8 ,uint8 ,uint8 ,uint8 ,uint16))
                                  (joy-axis-motion . (,uint8 ,uint8 ,uint8 ,int16))
                                  (joy-ball-motion . (,uint8 ,uint8 ,uint8 ,int16 ,int16))
                                  (joy-hat-motion . (,uint8 ,uint8 ,uint8 ,uint8))
                                  (joy-button-down . (,uint8 ,uint8 ,uint8 ,uint8))
                                  (joy-button-up . (,uint8 ,uint8 ,uint8 ,uint8))
                                  (video-resize . (,uint8 ,int ,int))))

;; (define event-field-names-alist '((active-event . (type gain state))
;;                                   (key-down . (type which state))
;;                                   (key-up . (type which state))
;;                                   (mouse-motion . (uint8 uint8 uint8 uint16 int16 int16))
;;                                   (mouse-button-down . (uint8 uint8 uint8 uint8 uint16))
;;                                   (mouse-button-up . (uint8 uint8 uint8 uint8 uint16))
;;                                   (joy-axis-motion . (uint8 uint8 uint8 int16))
;;                                   (joy-ball-motion . (uint8 uint8 uint8 int16 int16))
;;                                   (joy-hat-motion . (uint8 uint8 uint8 uint8))
;;                                   (joy-button-down . (uint8 uint8 uint8 uint8))
;;                                   (joy-button-up . (uint8 uint8 uint8 uint8))
;;                                   (video-resize . (uint8 int int))))
;; End of constants

;; ;; Event GOOPS declarations.
;; (define-class <event> ()
;;   (type #:getter get-type #:init-keyword #:type))

;; (define-class <keyboard-event> (<event>)
;;   (which #:getter get-which #:init-keyword #:which)
;;   (state #:getter get-state #:init-keyword #:which))
;; ;; End of event GOOPS declarations.

(define poll-event-primitive
  (pointer->procedure int
                      (dynamic-func "SDL_PollEvent" sdl-ob)
                      (list '*)))

(define poll-event
  (lambda (event-pointer)
    "SDL_PollEvent wrapper that returns #t or #f instead of 0 or 1."
    (= 1 (poll-event-primitive event-pointer))))

(define event-type
  (lambda (event-pointer)
    "Return the type field of the SDL_Event struct pointed to by EVENT-POINTER."
    (int->event-symbol (car (parse-c-struct event-pointer (list uint8))))))

(define int->event-symbol
  (lambda (event-int)
    "Return a symbol indicating the type of the (SDL) event represented by EVENT-INT."
    (vector-ref event-symbols event-int)))

(define event-pointer->vector
  (lambda (event-pointer)
    (let* ((event-field-types
           (assoc-ref event-field-types-alist
                      (event-type event-pointer)))
          (event-field-values #f))
      (if (eq? #f event-field-types)
          #f
          (begin
            (set! event-field-values
                  (parse-c-struct event-pointer
                                  event-field-types))
            (list->vector event-field-values))))))
